/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * © 2016 OX Software GmbH
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

/* global blankshield */

define('io.ox/office/baseframework/app/baseapplication',
    ['io.ox/core/extensions',
     'io.ox/core/notifications',
     'io.ox/files/api',
     'io.ox/office/tk/utils',
     'io.ox/office/tk/keycodes',
     'io.ox/office/tk/config',
     'io.ox/office/tk/io',
     'io.ox/office/tk/forms',
     'io.ox/office/tk/object/baseobject',
     'io.ox/office/tk/object/timermixin',
     'io.ox/office/baseframework/app/extensionregistry',
     'settings!io.ox/office',
     'gettext!io.ox/office/main'
    ], function (ext, Notifications, FilesAPI, Utils, KeyCodes, Config, IO, Forms, BaseObject, TimerMixin, ExtensionRegistry, Settings, gt) {

    'use strict';

    // private static functions ===============================================

    /**
     * Returns the document type from the passed application module name.
     * The document type equals the last directory name of the module name.
     *
     * @param {String} moduleName
     *  The application module name.
     *
     * @returns {String}
     *  The document type of the passed module name.
     */
    function getDocumentType(moduleName) {
        return moduleName.substr(moduleName.lastIndexOf('/') + 1);
    }

    /**
     * Returns the stand-alone mode from the server-side configuration flag
     * 'standalone', or the application's stand-alone URL hash fragment. This
     * function also checks if the application is loaded in an <iframe> and
     * gets the correct hash fragment from the it's source URL.
     *
     * @returns {Boolean}
     *  The state of the stand-alone mode.
     */
    function getStandaloneMode() {
        if (Config.getFlag('standalone')) { return true; }
        var standalone = (window.self !== window.top) ? _.deserialize(window.self.location.hash.substr(1)).standalone : _.url.hash('standalone');
        return standalone ? (standalone.toLowerCase() === 'true' || standalone === '1') : false;
    }

    /**
     * Returns the close timeout hash fragment from the application URL.
     * This value is limited to 600 and float values will be rounded up.
     *
     * @returns {Number|Null}
     *  The delay time after the application will be closed automatically, in
     *  milliseconds.
     */
    function getAutoCloseDelay() {
        var appCloseTimeout = _.url.hash('closetimer');
        if (!$.isNumeric(appCloseTimeout)) { return null; }
        return Math.min(600, Math.ceil(appCloseTimeout)) * 1000;
    }

    // class BaseApplication ==================================================

    /**
     * A mix-in class that defines common public methods for an application
     * that is based on a document file.
     *
     * @constructor
     *
     * @extends ox.ui.App
     * @extends TimerMixin
     *
     * @param {Function} ModelClass
     *  The constructor function of the document model class. MUST derive from
     *  the class Model. Receives a reference to this application instance.
     *  MUST NOT use the methods BaseApplication.getModel(),
     *  BaseApplication.getView(), or BaseApplication.getController() during
     *  construction. For further initialization depending on valid
     *  model/view/controller instances, the constructor can use the method
     *  BaseApplication.onInit().
     *
     * @param {Function} ViewClass
     *  The constructor function of the view class. MUST derive from the class
     *  View. Receives a reference to this application instance. MUST NOT use
     *  the methods BaseApplication.getModel(), BaseApplication.getView(), or
     *  BaseApplication.getController() during construction. For further
     *  initialization depending on valid model/view/controller instances, the
     *  constructor can use the method BaseApplication.onInit().
     *
     * @param {Function} ControllerClass
     *  The constructor function of the controller class. MUST derive from the
     *  class Controller. Receives a reference to this application instance.
     *  MUST NOT use the methods BaseApplication.getModel(),
     *  BaseApplication.getView(), or BaseApplication.getController() during
     *  construction. For further initialization depending on valid
     *  model/view/controller instances, the constructor can use the method
     *  BaseApplication.onInit().
     *
     * @param {Function} importHandler
     *  A function that will be called to import the document described in the
     *  file descriptor of this application. Will be called once after
     *  launching the application. Will be called in the context of this
     *  application. Must return a Deferred object that will be resolved or
     *  rejected after the document has been loaded.
     *
     * @param {Object} launchOptions
     *  All options passed to the core launcher (the ox.launch() method) that
     *  determine the actions to perform during application launch. The
     *  supported options are dependent on the actual application type. The
     *  following standard launch options are supported:
     *  @param {String} launchOptions.action
     *      Controls how to connect the application to a document file. If the
     *      launch options map is passed at all, this is the only mandatory
     *      launch option.
     *  @param {Object} [launchOptions.file]
     *      The descriptor of the file to be imported. May be missing for
     *      specific actions, e.g. when creating a new file.
     *
     * @param {Object} [appOptions]
     *  Static application options that have been passed to the static method
     *  BaseApplication.createLauncher().
     *
     * @param {Object} [initOptions]
     *  Optional parameters:
     *  @param {Function} [initOptions.flushHandler]
     *      A function that will be called before the document will be accessed
     *      at its source location, e.g. before downloading it, printing it, or
     *      sending it as e-mail. Will be called in the context of this
     *      application. Receives a string identifier as first parameter,
     *      specifying the cause of the flush request:
     *      - 'download': The document will be prepared for downloading or
     *          printing.
     *      - 'email': The document will be prepared to be sent as e-mail.
     *      May return a Deferred object. In this case, the application will
     *      wait until this Deferred object has been resolved or rejected. In
     *      the latter case, all subsequent actions will be skipped.
     */
    function BaseApplication(ModelClass, ViewClass, ControllerClass, importHandler, launchOptions, appOptions, initOptions) {

        var // self reference
            self = this,

            // the document model instance
            model = null,

            // application view: contains panes, tool bars, etc.
            view = null,

            // the controller instance as single connection point between model and view
            controller = null,

            // all destructor callbacks
            destructors = null,

            // all singleton values/objects for this application instance, mapped by key
            singletons = {},

            // file descriptor of the document edited by this application
            file = Utils.getObjectOption(launchOptions, 'file', null),

            // callback to flush the document before accessing its source
            flushHandler = Utils.getFunctionOption(initOptions, 'flushHandler', $.noop),

            // all registered before-quit handlers
            beforeQuitHandlers = [],

            // all registered quit handlers
            quitHandlers = [],

            // a Deferred object representing the initialization state (pending: early application construction, resolved: regular application runtime)
            initDef = $.Deferred(),

            // a Deferred object representing the import state (pending: importing, resolved/rejected: import finished with state)
            importDef = $.Deferred(),

            // all pending server requests currently running
            pendingRequests = [],

            // prevent further server requests after the quit handlers have been called
            requestsLocked = false,

            // time stamp when the application launcher has started
            launchStartTime = 0,

            // collector for timer log
            timerLogCollector = {},

            // application auto close timeout in seconds
            autoCloseDelay = getAutoCloseDelay(),

            // Deferred object of the current call to Application.quit()
            currentQuitDef = null,

            // identifier for a reason while quitting this application (passed to quit handlers)
            quitReason = null;

        // private methods ----------------------------------------------------

        /**
         * Updates the application title according to the current file name. If
         * the application does not contain a file descriptor, shows the
         * localized word 'Unnamed' as title.
         */
        function updateTitle() {

            var // the root node of the application window
                windowNode = self.getWindow().nodes.outer,
                // the title for the application launcher and browser window
                title = self.getShortFileName() || gt('unnamed');

            // set title for application launcher and browser tab/window
            self.setTitle(title);
            self.getWindow().setTitle(title);

            // add data from file descriptor to application root node, for Selenium testing
            if (file) {
                if (file.folder_id) { windowNode.attr('data-file-folder', file.folder_id); }
                if (file.id) { windowNode.attr('data-file-id', file.id); }
                if (file.filename) { windowNode.attr('data-file-name', file.filename); }
                if (file.version) { windowNode.attr('data-file-version', file.version); }
                if (file.source) { windowNode.attr('data-file-source', file.source); }
                if (file.attachment) { windowNode.attr('data-file-attachment', file.attachment); }
            }
        }

        /**
         * Updates the browser URL according to the current file descriptor.
         */
        function updateUrl() {
            // only for files from OX Drive
            if (file && !('source' in file) && self.isActive()) {
                self.setState({ folder: file.folder_id, id: file.id });
            }
        }

        /**
         * Logging time at specified key point.
         *
         * @param {String} key
         *  The key for the performance logger object.
         *
         * @param {Number} time
         *  The logged time.
         */
        function logTimer(key, time) {
            timerLogCollector[key] = time;
        }

        /**
         * Calls all handler functions contained in the passed array, and
         * returns a Deferred object that accumulates the result of all
         * handlers. Passes all additional function parameters to the passed
         * callback handlers.
         */
        function callHandlers(handlers) {

            var // arguments to be passed to the handlers
                args = _.toArray(arguments).slice(1),
                // execute all handlers and store their results in an array
                results = _.map(handlers, function (handler) { return handler.apply(self, args); });

            // accumulate all results into a single Deferred object
            return $.when.apply($, results);
        }

        /**
         * Handler for the 'unload' window events triggered when reloading or
         * closing the entire page. Executes the quit handlers, but not the
         * before-quit handlers.
         */
        function unloadHandler() {
            self.executeQuitHandlers('unload');
        }

        /**
         * Downloads the document file currently opened by this application.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {String} [options.convert]
         *      If specified, the document will be converted to the requested
         *      file format (for example, the value 'pdf' will convert the
         *      document to PDF). If omitted, the document will be downloaded
         *      in its native file format.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved if the
         *  document file has been downloaded successfully, or rejected
         *  otherwise.
         */
        function download(options) {

            // creates and returns a rejected promise, containing the specified error cause
            function reject(cause) {
                return $.Deferred().reject({ cause: cause });
            }

            // nothing to do, if no file descriptor is available (internal error, server error, etc.)
            if (!self.hasFileDescriptor()) { return reject('missingfile'); }

            // document import must have been finished successfully
            if (!self.isImportSucceeded()) { return reject('importfailed'); }

            var // the promise that will be resolved with the download URL
                downloadPromise = null,
                // the pop-up window used to download the file
                popupWindow = null,
                // the target file format
                fileFormat = Utils.getStringOption(options, 'convert', 'native');

            // open pop-up window early to prevent browser pop-up blockers (bug 28251)
            popupWindow = blankshield.open('about:blank', '_blank');
            window.focus();

            // browser plug-ins may still prevent opening pop-up windows
            if (!popupWindow) { return reject('popupblocker'); }

            // block application window while waiting for the file URL
            view.enterBusy({ immediate: true });

            // first, call the flush handler passed to the application constructor
            downloadPromise = $.when(flushHandler.call(self, 'download'));

            // check availability of document converter, if file will be converted to another format
            if (fileFormat !== 'native') {

                // query available server features
                downloadPromise = downloadPromise.then(function () {
                    return self.sendFeatureRequest();
                });

                // check for document converter
                downloadPromise = downloadPromise.then(function (availableFeatures) {
                    if (!_.contains(availableFeatures, 'documentconverter')) {
                        return reject('docconverter');
                    }
                });
            }

            // propagate changed file to the files API (bug 30246: not for mail/task attachments!)
            downloadPromise = downloadPromise.then(function () {
                if (!('source' in file)) {
                    return FilesAPI.propagate('change', file);
                }
            });

            // download the file in the pop-up window opened before
            downloadPromise = downloadPromise.then(function () {

                var // additional parameters inserted into the URL
                    params = {
                        action: 'getdocument',
                        documentformat: fileFormat,
                        mimetype: file.file_mimetype ? encodeURIComponent(file.file_mimetype) : '',
                        nocache: _.uniqueId() // needed to trick the browser cache (is not evaluated by the backend)
                    };

                // set the resulting file URL at the pop-up window (bug 36734: use current version)
                popupWindow.location = self.getServerModuleUrl(IO.CONVERTER_MODULE_NAME, params, { currentVersion: true });

                // show/focus the pop-up window
                popupWindow.focus();
            });

            // hide the pop-up window, if URL is not available
            downloadPromise.fail(function () {
                popupWindow.close();
                view.grabFocus();
            });

            // leave busy mode after download
            downloadPromise.always(function () {
                view.leaveBusy();
            });

            return downloadPromise;
        }

        // public methods -----------------------------------------------------

        /**
         * Returns the document type supported by this application. The
         * document type equals the directory name of the application.
         *
         * @returns {String}
         *  The document type of this application.
         */
        this.getDocumentType = function () {
            return getDocumentType(this.getName());
        };

        /**
         * Returns the document model instance of this application.
         *
         * @returns {DocumentModel}
         *  The document model instance of this application.
         */
        this.getModel = function () {
            return model;
        };

        /**
         * Returns the view instance of this application.
         *
         * @returns {View}
         *  The view instance of this application.
         */
        this.getView = function () {
            return view;
        };

        /**
         * Returns the controller instance of this application.
         *
         * @returns {Controller}
         *  The controller instance of this application.
         */
        this.getController = function () {
            return controller;
        };

        /**
         * Invokes the passed callback function as soon as this application has
         * finished its early construction phase. During early construction,
         * the model, view, and controller instances may not be available yet.
         * If this method is used after early construction, the passed callback
         * function will be invoked immediately.
         *
         * @attention
         *  Should be used in all class initialization code relying on existing
         *  model, view, and controller instances.
         *
         * @param {Function} callback
         *  The callback function invoked after the application has finished
         *  its early initialization.
         *
         * @param {Object} [context]
         *  The context bound to the invoked callback function.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.onInit = function (callback, context) {
            initDef.done(_.bind(callback, context));
            return this;
        };

        /**
         * Registers a destructor callback function that will be invoked when
         * the application shuts down.
         *
         * @attention
         *  Intended to be used by the internal implementations of sub classes.
         *  DO NOT CALL from external code!
         *
         * @param {Function} destructor
         *  A destructor callback function. During shut-down, all registered
         *  destructor callbacks will be invoked in reverse order of
         *  registration. The callback will be invoked in the context of this
         *  application.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.registerDestructor = function (destructor) {
            (destructors || (destructors = [])).unshift(destructor);
            return this;
        };

        /**
         * Returns the global user settings for all applications of the same
         * type.
         *
         * @param {String} key
         *  The unique key of the user setting.
         *
         * @param defValue
         *  The default value in case the user setting does not exist yet.
         *
         * @returns
         *  The value of the global user setting.
         */
        this.getUserSettingsValue = function (key, defValue) {
            // for debugging purpose we output the whole settings object
            //Utils.log(JSON.stringify(Settings.get()));
            return Settings.get(this.getDocumentType() + '/' + key, defValue);
        };

        /**
         * Changes a global user setting for all applications of the same type.
         *
         * @param {String} key
         *  The unique key of the user setting.
         *
         * @param value
         *  The new value of the user setting.
         *
         * @returns {BaseApplication}
         *  A reference to this application instance.
         */
        this.setUserSettingsValue = function (key, value) {
            Settings.set(this.getDocumentType() + '/' + key, value).save();
            return this;
        };

        /**
         * Returns a lazy-initialized application-wide value or object. If the
         * singleton has not been created yet, calls the passed initialization
         * callback function and caches its return value internally.
         *
         * @param {String} key
         *  The unique key of the singleton.
         *
         * @param {Function} initHandler
         *  The initialization callback function that will be called when the
         *  singleton will be queried the first time for this application. Will
         *  be called in the context of this application instance. Receives the
         *  key of the singleton as first parameter (so that the same callback
         *  function can be used to initialize different singletons).
         *
         * @returns {Any}
         *  The singleton value.
         */
        this.getOrCreateSingleton = function (key, initHandler) {
            return (key in singletons) ? singletons[key] : (singletons[key] = initHandler.call(this, key));
        };

        /**
         * Returns whether this application has been launched in stand-alone
         * mode.
         *
         * @returns {Boolean}
         *  Whether this application has been launched in stand-alone mode.
         */
        this.isStandalone = function () {
            return getStandaloneMode();
        };

        /**
         * Returns whether this application is currently active in OX AppSuite.
         *
         * @returns {Boolean}
         *  Whether this application is currently active in OX AppSuite
         */
        this.isActive = function () {
            return !!this.getWindow().state.visible;
        };

        /**
         * Returns the unique DOM identifier of the outer root node of this
         * application.
         *
         * @returns {String}
         *  The unique window DOM identifier.
         */
        this.getWindowId = function () {
            return this.getWindow().nodes.outer.attr('id');
        };


        /**
         * Returns the absolute start time of the launcher.
         *
         * @returns {Number}
         *  The launcher start time.
         */
        this.getLaunchStartTime = function () {
            return launchStartTime;
        };

        /**
         * Returns the collector for all timers.
         *
         * @returns {Object}
         *  The collector for timer logs.
         */
        this.getTimerLogCollector = function () {
            return timerLogCollector;
        };

        // file descriptor ----------------------------------------------------

        /**
         * Returns whether this application contains a valid file descriptor.
         */
        this.hasFileDescriptor = function () {
            return _.isObject(file);
        };

        /**
         * Returns a clone of the file descriptor of the document edited by
         * this application.
         *
         * @returns {Object}
         *  A clone of the current file descriptor.
         */
        this.getFileDescriptor = function () {
            return _.clone(file);
        };

        /**
         * Registers the file descriptor of the new document created by this
         * application. Must not be called if the application already contains
         * a valid file descriptor.
         *
         * @param {Object} newFile
         *  The file descriptor of the new file created by the caller of this
         *  method.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.registerFileDescriptor = function (newFile) {
            if (_.isObject(file)) {
                Utils.error('BaseApplication.registerFileDescriptor(): file descriptor exists already');
            } else {
                file = newFile;
                FilesAPI.propagate('new', file);
                updateTitle();
                updateUrl();
            }
            return this;
        };

        /**
         * Updates the current file descriptor of the document edited by this
         * application.
         *
         * @param {Object} fileOptions
         *  All file descriptor properties to be updated.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.updateFileDescriptor = function (fileOptions) {
            if (_.isObject(file)) {
                _.extend(file, fileOptions);
                FilesAPI.propagate('change', file);
                updateTitle();
                updateUrl();
            }
            return this;
        };

        /**
         * Returns an object with attributes describing the file currently
         * opened by this application.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.encodeUrl=false]
         *      If set to true, special characters not allowed in URLs will be
         *      encoded.
         *  @param {Boolean} [options.currentVersion=false]
         *      If set to true, the version stored in the file descriptor will
         *      NOT be inserted into the result (thus, the server will always
         *      access the current version of the document).
         *
         * @returns {Object|Null}
         *  An object with file attributes, if existing; otherwise null.
         */
        this.getFileParameters = function (options) {

            var // function to encode a string to be URI conforming if specified
                encodeString = Utils.getBooleanOption(options, 'encodeUrl', false) ? encodeURIComponent : _.identity,
                // the resulting file parameters
                parameters = null;

            if (file) {
                parameters = {};

                // add the parameters to the result object, if they exist in the file descriptor
                _.each(['id', 'folder_id', 'filename', 'version', 'source', 'attached', 'module'], function (name) {
                    if (_.isString(file[name])) {
                        parameters[name] = encodeString(file[name]);
                    } else if (_.isNumber(file[name])) {
                        parameters[name] = file[name];
                    }
                });

                // remove the version identifier, if specified
                if (Utils.getBooleanOption(options, 'currentVersion', false)) {
                    delete parameters.version;
                }
            }

            return parameters;
        };

        /**
         * Returns the full file name of the current file, with file extension.
         *
         * @returns {String|Null}
         *  The file name of the current file descriptor; or null, if no file
         *  descriptor exists.
         */
        this.getFullFileName = function () {
            return Utils.getStringOption(file, 'filename', null);
        };

        /**
         * Returns the short file name of the current file (without file
         * extension).
         *
         * @returns {String|Null}
         *  The short file name of the current file descriptor; or null, if no
         *  file descriptor exists.
         */
        this.getShortFileName = function () {
            var fileName = this.getFullFileName();
            return _.isString(fileName) ? ExtensionRegistry.getBaseName(fileName) : null;
        };

        /**
         * Returns the file extension of the current file (without leading
         * period).
         *
         * @returns {String|Null}
         *  The file extension of the current file descriptor; or null, if no
         *  file descriptor exists.
         */
        this.getFileExtension = function () {
            var fileName = this.getFullFileName();
            return _.isString(fileName) ? ExtensionRegistry.getExtension(fileName) : null;
        };

        /**
         * Returns the import promise that is pending as long as the document
         * will be imported, and that will be resolved after successful import,
         * or rejected after import has failed.
         */
        this.getImportPromise = function () {
            return importDef.promise();
        };

        /**
         * Return whether importing the document has been completed regardless
         * whether it was successful. Will be false while the import promise is
         * pending, and true after the import promise has been resolved or
         * rejected.
         *
         * @returns {jQuery.Promise}
         *  The import promise of this document.
         */
        this.isImportFinished = function () {
            return importDef.state() !== 'pending';
        };

        /**
         * Return whether importing the document was successful. Will be false
         * while the import promise is pending or after it has been rejected,
         * and true after the import promise has been resolved.
         *
         * @returns {Boolean}
         *  Whether importing the document has been completed successfully.
         */
        this.isImportSucceeded = function () {
            return importDef.state() === 'resolved';
        };

        /**
         * Return whether importing the document has failed. Will be false
         * while the import promise is pending or after it has been resolved,
         * and true after the import promise has been rejected.
         *
         * @returns {Boolean}
         *  Whether importing the document is finished and has failed.
         */
        this.isImportFailed = function () {
            return importDef.state() === 'rejected';
        };

        /**
         * Invokes the passed callback function as soon as this application has
         * finished importing the document. If this method is used after
         * importing the document, the passed callback function will be invoked
         * immediately. This method is a convenience shortcut for using the
         * code line 'app.getImportPromise().always(callback);'.
         *
         * @param {Function} callback
         *  The callback function invoked after the application has finished
         *  importing the document.
         *
         * @param {Object} [context]
         *  The context bound to the invoked callback function.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.onImport = function (callback, context) {
            importDef.always(_.bind(callback, context));
            return this;
        };

        /**
         * Invokes the passed callback function as soon as this application has
         * finished importing the document successfully. If this method is used
         * after importing the document, the passed callback function will be
         * invoked immediately (unless import has failed). This method is a
         * convenience shortcut for using the code line
         * 'app.getImportPromise().done(callback);'.
         *
         * @param {Function} callback
         *  The callback function invoked after the application has finished
         *  importing the document successfully.
         *
         * @param {Object} [context]
         *  The context bound to the invoked callback function.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.onImportSuccess = function (callback, context) {
            importDef.done(_.bind(callback, context));
            return this;
        };

        /**
         * Invokes the passed callback function as soon as this application has
         * failed importing the document. If this method is used after
         * importing the document, the passed callback function will be invoked
         * immediately (unless import has succeeded). This method is a
         * convenience shortcut for using the code line
         * 'app.getImportPromise().fail(callback);'.
         *
         * @param {Function} callback
         *  The callback function invoked after the application has failed
         *  importing the document.
         *
         * @param {Object} [context]
         *  The context bound to the invoked callback function.
         *
         * @returns {BaseApplication}
         *  A reference to this instance.
         */
        this.onImportFailure = function (callback, context) {
            importDef.fail(_.bind(callback, context));
            return this;
        };

        /**
         * Returns whether the application contains pending document changes
         * not yet sent to the server. Intended to be overwritten on demand by
         * sub classes.
         *
         * @attention
         *  Do not rename this method. The OX core uses it to decide whether to
         *  show a warning before the browser refreshes or closes the page.
         *
         * @returns {Boolean}
         *  Whether the application contains pending document changes.
         */
        this.hasUnsavedChanges = function () {
            return false;
        };

        /**
         * Checks if application is processing before-quit or quit handlers.
         *
         * @returns {Boolean}
         *  Whether the application is currently processing any before-quit or
         *  quit handlers.
         */
        this.isInQuit = function () {
            return currentQuitDef !== null;
        };

        // server requests ----------------------------------------------------

        /**
         * Sends a request to the server and returns the promise of a Deferred
         * object waiting for the response. The unique identifier of this
         * application will be added to the request parameters automatically.
         * See method IO.sendRequest() for further details.
         *
         * @param {String} module
         *  The name of the server module that will receive the request.
         *
         * @param {Object} [params]
         *  Parameters that will be inserted into the request URL (method GET),
         *  or into the request body (method POST).
         *
         * @param {Object} [options]
         *  Optional parameters. See method IO.sendRequest() for details.
         *
         * @returns {jQuery.Promise}
         *  The Promise of the request. See method IO.sendRequest() for
         *  details.
         */
        this.sendRequest = function (module, params, options) {

            var // the AJAX request, as (abortable) jQuery promise
                request = null;

            // do not allow to send server requests after the quit handlers have been called
            if (requestsLocked) {
                return $.Deferred().reject('quit');
            }

            // add the application UID to the request parameters
            params = _.extend({}, params, { uid: this.get('uniqueID') });

            // send the request
            request = IO.sendRequest(module, params, options);

            // store the request internally for automatic abort on application quit
            pendingRequests.push(request);
            request.always(function () {
                pendingRequests = _.without(pendingRequests, request);
            });

            return request;
        };

        /**
         * Sends a request to the document filter server module to determine
         * if certain server features are working.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object. On success of the request, the
         *  promise will be resolved with an array of strings containing all
         *  supported server features.
         */
        this.sendFeatureRequest = function () {
            return this.sendRequest(IO.FILTER_MODULE_NAME, { action: 'getfeatures' }, {
                resultFilter: function (data) {
                    var features = Utils.getStringOption(data, 'features');
                    return _.isString(features) ? features.split(/,/) : undefined;
                }
            });
        };

        /**
         * Sends a request to the server and returns the Promise of a Deferred
         * object waiting for the response. The unique identifier of this
         * application, and the parameters of the file currently opened by the
         * application will be added to the request parameters automatically.
         * See method IO.sendRequest() for further details.
         *
         * @param {String} module
         *  The name of the server module that will receive the request.
         *
         * @param {Object} [params]
         *  Parameters that will be inserted into the request URL (method GET),
         *  or into the request body (method POST).
         *
         * @param {Object} [options]
         *  Optional parameters. See method IO.sendRequest() for details.
         *
         * @returns {jQuery.Promise}
         *  The Promise of the request. Will be rejected immediately, if this
         *  application is not connected to a document file. See method
         *  IO.sendRequest() for details.
         */
        this.sendFileRequest = function (module, params, options) {

            // reject immediately if no file is present
            if (!this.hasFileDescriptor()) {
                return $.Deferred().reject();
            }

            // extend parameters with file settings
            params = _.extend({}, params, this.getFileParameters());

            // send the request
            return this.sendRequest(module, params, options);
        };

        /**
         * Creates and returns the URL of a server request.
         *
         * @param {String} module
         *  The name of the server module.
         *
         * @param {Object} [params]
         *  Additional parameters inserted into the URL.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Boolean} [options.currentVersion=false]
         *      If set to true, the version stored in the file descriptor will
         *      NOT be inserted into the generated URL (thus, the server will
         *      always access the current version of the document).
         *
         * @returns {String|Undefined}
         *  The final URL of the server request; or undefined, if the
         *  application is not connected to a document file, or the current
         *  session is invalid.
         */
        this.getServerModuleUrl = function (module, params, options) {

            // return nothing if no file is present
            if (!ox.session || !this.hasFileDescriptor()) {
                return;
            }

            var // the parameters for the file currently loaded
                fileParams = this.getFileParameters(Utils.extendOptions(options, { encodeUrl: true }));

            // add default parameters (session and UID), and file parameters
            params = _.extend({ session: ox.session, uid: this.get('uniqueID') }, fileParams, params);

            // build and return the resulting URL
            return ox.apiRoot + '/' + module + '?' + _.map(params, function (value, name) { return name + '=' + value; }).join('&');
        };

        /**
         * Creates an <img> element, sets the passed URL, and returns a
         * Deferred object waiting that the image has been loaded. If the
         * browser does not trigger the appropriate events after the specified
         * timeout, the Deferred object will be rejected automatically. If the
         * application has been shut down before the image has been loaded, the
         * pending Deferred object will be rejected automatically.
         *
         * @param {String} url
         *  The image URL.
         *
         * @param {Object} [options]
         *  Optional parameters:
         *  @param {Number} [options.timeout=10000]
         *      The time before the Deferred will be rejected without response
         *      from the image node.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved with the
         *  image element (as jQuery object), when the 'load' event has been
         *  received from the image node; or rejected, if the 'error' event has
         *  been received from the image, or after the specified timeout delay
         *  without response.
         */
        this.createImageNode = function (url, options) {

            var // the result Deferred object
                def = $.Deferred(),
                // the image node
                imgNode = $('<img>', { src: url }),
                // the duration of the failure timeout
                timeout = Utils.getIntegerOption(options, 'timeout', 10000, 0),
                // the resulting promise
                promise = this.createAbortablePromise(def, $.noop, timeout),
                // the event handlers for the image node
                handlers = {
                    load: function () { def.resolve(imgNode); },
                    error: function () { def.reject(); }
                };

            // wait that the image is loaded
            imgNode.one(handlers);
            return promise.always(function () { imgNode.off(handlers); });
        };

        /**
         * Safely destroys the passed image nodes. On platforms with restricted
         * memory (especially iPad and iPhone), allocating too many images may
         * result in an immediate browser crash. Simply removing image nodes
         * from the DOM does not free the image resource data allocated by the
         * browser. The suggested workaround is to replace the 'src' attribute
         * of the image node with a local small image which will result in
         * releasing the original image data by the internal resource manager.
         * To be really sure that the garbage collector does not destroy the
         * DOM node instance before the resource manager releases the image
         * data, a reference to the node will be kept in memory for additional
         * 60 seconds.
         *
         * @param {HTMLElement|jQuery} nodes
         *  The image elements, or any other container elements with descendant
         *  image nodes, either as single DOM node, or as jQuery collection
         *  with one or multiple elements.
         */
        this.destroyImageNodes = function (nodes) {

            var // filter passed image nodes, find descendant image nodes
                imgNodes = $(nodes).filter('img').add($(nodes).find('img'));

            if (imgNodes.length > 0) {
                Utils.insertHiddenNodes(imgNodes);
                $(imgNodes).off().attr('src', 'data:image/gif;base64,R0lGODdhAgACAIAAAAAAAP///ywAAAAAAgACAAACAoRRADs=');
                _.delay(function () { imgNodes.remove(); imgNodes = null; }, 60000);
            }
        };

        // application setup --------------------------------------------------

        /**
         * Registers a quit handler function that will be executed before the
         * application will be closed. The callback handler function may veto
         * the quit request to keep the application alive.
         *
         * @param {Function} beforeQuitHandler
         *  A function that will be called before the application will be
         *  closed. Will be called in the context of this application instance.
         *  Receives the reason for quitting that has been passed to the method
         *  BaseApplication.quit() as first parameter. May return a Deferred
         *  object, which must be resolved or rejected by the quit handler
         *  function. If the Deferred object will be rejected, the application
         *  remains alive.
         *
         * @returns {BaseApplication}
         *  A reference to this application instance.
         */
        this.registerBeforeQuitHandler = function (beforeQuitHandler) {
            beforeQuitHandlers.push(beforeQuitHandler);
            return this;
        };

        /**
         * Registers a quit handler function that will be executed before the
         * application will be destroyed. This may happen if the application
         * has been closed normally, if the user logs out from the entire
         * AppSuite, or before the browser window or browser tab will be closed
         * or refreshed. If the quit handlers return a Deferred object, closing
         * the application, and logging out will be deferred until the Deferred
         * objects have been resolved or rejected.
         *
         * @param {Function} quitHandler
         *  A function that will be called before the application will be
         *  closed. Receives the reason for quitting that has been passed to
         *  the method BaseApplication.quit() as first parameter. Will be
         *  called in the context of this application instance. May return a
         *  Deferred object, which must be resolved or rejected by the quit
         *  handler function.
         *
         * @returns {BaseApplication}
         *  A reference to this application instance.
         */
        this.registerQuitHandler = function (quitHandler) {
            quitHandlers.push(quitHandler);
            return this;
        };


        /**
         *  registers show & hide handler to the appwindow of this app
         *  the visible state get assigned to the handler ab boolean
         *  handler is called directly once with the current state
         *
         *  @param {Function} stateHandler
         */
        this.registerVisibleStateHandler = function(stateHandler) {
            var wnd = self.getWindow();
            wnd.on('show', function() {
                stateHandler(true);
            });
            wnd.on('hide', function() {
                stateHandler(false);
            });
            stateHandler(wnd.state.visible);
        };

        /**
         * Executes all registered quit handlers and returns a Deferred object
         * that will be resolved or rejected according to the results of the
         * handlers.
         *
         * @internal
         *  Not for external use. Only used from the implementation of the
         *  'io.ox/core/logout' extension point and the window's 'unload' event
         *  handler.
         *
         * @returns {jQuery.Promise}
         *  A promise that will be resolved/rejected when all quit handlers
         *  have been executed.
         */
        this.executeQuitHandlers = function (reason) {
            // Bug 28044: unbind the 'unload' event handler immediately, otherwise
            // logging out may trigger quit handlers twice (from logout and unload)
            $(window).off('unload', unloadHandler);
            // call quit handlers (may send server requests)
            var promise = callHandlers(quitHandlers, reason);
            // immediately prevent sending other server requests afterwards
            requestsLocked = true;
            return promise;
        };

        // application runtime ------------------------------------------------

        /**
         * Downloads the document file currently opened by this application.
         * This method should be used by the UI as the method will provide
         * error messages to the user in case of an error.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved if the
         *  document file has been downloaded successfully, or rejected
         *  otherwise.
         */
        this.download = function () {

            var // promise for the converted and downloaded document
                downloadPromise = download();

            // show error message on failure
            downloadPromise.fail(function (response) {

                var // the alert type
                    type = 'error',
                    // the message text
                    message = null;

                switch (Utils.getStringOption(response, 'cause')) {
                case 'importfailed':
                    message = gt('The document was not loaded successfully. Download is not possible.');
                    break;
                case 'popupblocker':
                    type = 'warning';
                    message = gt('The browser has blocked a pop-up window. Please enable pop-ups for download.');
                    break;
                default:
                    message = gt('The server does not answer the request.');
                }

                view.yell({ type: type, headline: gt('Download Error'), message: message });
            });

            return downloadPromise;
        };

        /**
         * Prints the document file currently opened by this application.
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved if the
         *  document file has been printed successfully, or rejected otherwise.
         */
        this.print = function () {

            var // promise for the converted and downloaded document
                downloadPromise = download({ convert: 'pdf' });

            // show error message on failure
            downloadPromise.fail(function (response) {

                var // the alert type
                    type = 'error',
                    // the message text
                    message = null;

                switch (Utils.getStringOption(response, 'cause')) {
                case 'docconverter':
                    message = gt('The document converter is currently not available.');
                    break;
                case 'importfailed':
                    message = gt('The document was not loaded successfully. Printing is not possible.');
                    break;
                case 'popupblocker':
                    type = 'warning';
                    message = gt('The browser has blocked a pop-up window. Please enable pop-ups for printing.');
                    break;
                default:
                    message = gt('The server does not answer the request.');
                }

                view.yell({ type: type, headline: gt('Printing Error'), message: message });
            });

            return downloadPromise;
        };

        /**
         * Creates a new mail message with the current document attached
         *
         * @returns {jQuery.Promise}
         *  The Promise of a Deferred object that will be resolved if the mail
         *  application has been launched successfully, or rejected otherwise.
         */
        this.sendMail = function () {

            var // the Promise with the result of the flush callback handler
                flushPromise = (this.isImportSucceeded() && this.hasFileDescriptor()) ? $.when(flushHandler.call(this, 'email')) : $.Deferred().reject();

            // launch Mail application if flushing was successful
            flushPromise = flushPromise.then(function () {
                return ox.launch('io.ox/mail/write/main').then(function () {

                    var // the mail writer application
                        mailApp = this;

                    // initializes the mail writer application responding to the specified mail
                    function replyToMail(mail) {
                        return mailApp.replyall({ folder: mail.folder_id, id: mail.id });
                    }

                    // appends the current document (must be a Drive file) to the mail writer application
                    function appendDriveFile() {
                        mailApp.addFiles([{
                            filename: file.filename,
                            folder_id: file.folder_id,
                            id: file.id,
                            size: file.file_size,
                            file_size: file.file_size
                        }], 'infostore');
                    }

                    // bug 32760: special handling, if file is an attachment, e.g. of a mail or task
                    if ('source' in file) {
                        switch (file.source) {

                        case 'mail':
                            return replyToMail(file).done(function () {
                                mailApp.setAttachments({ attachments: [{
                                    atmsgref: file.folder_id + '/' + file.id,
                                    content: null,
                                    content_type: file.file_mimetype,
                                    disp: 'attachment',
                                    filename: file.filename,
                                    group: 'attachment',
                                    id: file.attached,
                                    size: file.file_size,
                                    type: self.getFileExtension()
                                }] });
                            });

                        case 'task':
                            Utils.warn('BaseApplication.sendMail(): attaching a task attachment to a mail not implemented');
                            return mailApp.compose();

                        default:
                            Utils.warn('BaseApplication.sendMail(): unknown file source: "' + file.source + '"');
                            return mailApp.compose();
                        }
                    }

                    // bug 34300: mail attachments copied to Drive
                    if (_.isObject(file.origin)) {
                        if (file.origin.source === 'mail') {
                            return replyToMail(file.origin).done(appendDriveFile);
                        }
                        Utils.warn('BaseApplication.sendMail(): unknown file origin: "' + file.origin.source + '"');
                        return mailApp.compose();
                    }

                    // default case: attach a file from OX Drive
                    return mailApp.compose().done(appendDriveFile);
                });
            }, function () {

                var // the message text
                    message = null,
                    // cause
                    cause = self.isImportSucceeded() ? 'unknown' : 'importfailed';

                switch (cause) {
                    case 'importfailed':
                        message = gt('The document was not loaded successfully. Send as mail is not possible.');
                        break;
                    default:
                        message = gt('The server does not answer the request. Send as mail is not possible.');
                }

                view.yell({ type: 'error', headline: gt('Send as mail'), message: message });
            });

            return flushPromise;
        };

        /**
         * Closes this application instance. First, calls the registered
         * before-quit handlers. If none of the handlers vetoes (e.g. due to
         * unsaved changes, or after asking the user), calls the registered
         * quit handlers, and destroys the MVC instances and this application.
         *
         * @attention
         *  After calling this method, this application instance may be
         *  destroyed. An external caller MUST NOT do anything else with this
         *  instance.
         *
         * @param {String} [reason='user']
         *  The reason for quitting the application. The reason will be passed
         *  to all registered before-quit handlers, and all registered quit
         *  handlers. The default reason 'user' is intended for quitting the
         *  application due to a user interaction, e.g. pressing a Quit button.
         *
         * @returns {jQuery.Promise}
         *  The promise of a Deferred object that will be resolved when the
         *  application has actually been closed (no before-quit handler has
         *  vetoed quitting).
         */
        this.quit = (function (coreQuitMethod) {
            return function (reason) {

                var // remember current quit() method (needs to be restored after resuming,
                    // derived methods may have added their own quit() method)
                    origQuitMethod = this.quit;

                // create the result deferred (rejecting means resume application)
                currentQuitDef = $.Deferred();

                // rejected: resume to running state of application
                currentQuitDef.fail(function () {
                    self.quit = origQuitMethod;
                    currentQuitDef = null;
                });

                // override 'quit()' method to prevent repeated/recursive calls while in quit mode
                this.quit = function () { return currentQuitDef.promise(); };

                // store the reason in an internal variable (it is not possible
                // to pass any data to the core quit() method)
                quitReason = _.isString(reason) ? reason : 'user';

                // ensure that initialization is complete before calling destructors
                initDef.always(function () {

                    // Call core quit() method deferred to prevent fast destruction of any application
                    // child instances while calling own quit(), the core method will invoke the
                    // callback function registered via setQuit() that actually calls the registered
                    // quit handlers, and destroys the application (unless quit has been canceled).
                    _.defer(coreQuitMethod);
                });

                return currentQuitDef.promise();
            };
        }(_.bind(this.quit, this)));

        // initialization -----------------------------------------------------

        // add timer methods (needs the registerDestructor() method)
        TimerMixin.call(this);

        // set launcher callback invoked from base class
        this.setLauncher(function () {

            var // the application window
                win = null,
                // the value of the 'spellcheck' attribute of the <body> element
                spellState = null,
                // duration of a single step between two log messages
                lapTime = _.now(),
                // launch source
                actionSource = Utils.getStringOption(launchOptions, 'actionSource', null);

            function logLaunchDuration(message, key) {
                var time = _.now();
                if (key) { logTimer(key, time); }
                Utils.info('BaseApplication.setLauncher(): ' + message + ' finished in ' + (time - lapTime) + 'ms (total ' + (time - launchStartTime) + 'ms)');
                lapTime = time;
            }

            Utils.info('BaseApplication.setLauncher(): starting construction');
            launchStartTime = lapTime;

            // check that launch options have been passed
            if (!_.isObject(launchOptions) || !_.isString(launchOptions.action)) {
                return $.Deferred().reject();
            }

            // create the application window
            win = ox.ui.createWindow({
                name: self.getName(),
                search: Utils.getBooleanOption(appOptions, 'search', false),
                chromeless: Utils.getBooleanOption(appOptions, 'chromeless', false)
            });

            // set the window at the application instance
            self.setWindow(win);
            updateTitle();

            // wait for unload events and execute quit handlers
            $(window).on('unload', unloadHandler);

            // create the MVC instances
            model = new ModelClass(self);
            view = new ViewClass(self);
            controller = new ControllerClass(self);

            // disable global spell checking while this application is active
            win.on('show', function () {
                spellState = $(document.body).attr('spellcheck');
                $(document.body).attr('spellcheck', false);
            });
            win.on('beforehide', function () {
                if (spellState) {
                    $(document.body).attr('spellcheck', spellState);
                } else {
                    $(document.body).removeAttr('spellcheck');
                }
            });

            // bug 29711: update URL when applications becomes active
            win.on('show', updateUrl);

            // launch application per URL
            if (actionSource === 'url') {
                var fail = function(message) {
                    var cause = Utils.getStringOption(message, 'error', 'unknown');
                    Utils.error('BaseApplication.setLauncher(): Failed loading file descriptor from Files API. Error code: ' + cause);
                    initDef.reject(message);
                };
                if (launchOptions.template) {
                    FilesAPI.get(launchOptions.templateFile).done(function (fileDescriptor) {
                        // get complete file descriptor asynchronously and save to local templateFile
                        launchOptions.templateFile = fileDescriptor;
                        // resolve app initialization and continue app launching
                        initDef.resolve();
                    }).fail(fail);
                } else {
                    FilesAPI.get(launchOptions.file).done(function (fileDescriptor) {
                        // get complete file descriptor asynchronously and save to local file descriptor
                        file = fileDescriptor;
                        // resolve app initialization and continue app launching
                        initDef.resolve();
                    }).fail(fail);
                }
            } else { // normal launching with the given file descriptor from launchOptions
                initDef.resolve();
            }

            // call initialization handlers before showing the application window
            // in order to get the 'open' event of the window at all, it must be shown (also without file).
            // in case of URL launching, wait until the complete file descriptor is loaded successfully from FilesAPI
            initDef.done(function () {

                win.show(function () {

                    // log the time needed to construct the application window and MVC instances
                    logLaunchDuration('application construction', 'appConstructed');

                    var // the resulting promise for entire GUI initialization (no-op timer used to get a promise that simply
                        // resolves deferred, needed to run all code registered with app.onInit() before initializing the GUI)
                        initGuiPromise = self.executeDelayed($.noop);

                    // hide the view, call the view initialization handler
                    initGuiPromise = initGuiPromise.then(function () {
                        return view.hideBeforeImport().always(function () {
                            logLaunchDuration('GUI preparation', 'guiPrepared');
                        });
                    });

                    // call the import handler
                    initGuiPromise = initGuiPromise.then(function () {
                        return importHandler.call(self).always(function () {
                            logLaunchDuration('document import', 'documentImported');
                        });
                    });

                    // makes the view visible after import
                    function showAfterImport(resolved, result) {

                        // nothing to do, if application is shutting down
                        if (self.isInQuit()) { return result; }

                        var // the Deferred object used to forward the original state and result
                            def = $.Deferred();

                        // make the view visible
                        view.showAfterImport().always(function () {
                            self.getWindow().nodes.outer.attr('data-doc-loaded', true);
                            updateTitle();
                            logLaunchDuration('GUI initialization', 'guiInitialized');
                            // forward state and result of the originating promise
                            if (resolved) { def.resolve(result); } else { def.reject(result); }
                        });

                        return def.promise();
                    }

                    // finally, make the view visible (also if anything before has failed)
                    initGuiPromise = initGuiPromise.then(_.partial(showAfterImport, true), _.partial(showAfterImport, false));

                    // resolve the global import Deferred object on success, unless the
                    // application is already closing (import canceled)
                    initGuiPromise.done(function (result) {
                        if (!self.isInQuit()) {
                            // set app auto-close timeout, if specified
                            if (_.isNumber(autoCloseDelay)) {
                                self.executeDelayed(function () { self.quit('autoclose'); }, autoCloseDelay);
                            }
                            importDef.resolve(result);
                        }
                    });

                    // handle import failures, unless the  application is already closing (import canceled)
                    initGuiPromise.fail(function (result) {
                        var cause = Utils.getStringOption(result, 'cause', 'unknown');
                        Utils.error('BaseApplication.launch(): Importing document "' + self.getFullFileName() + '" failed. Error code: "' + cause + '".');
                        // do not show any messages nor trigger events, if application is shutting down
                        if (!self.isInQuit()) { importDef.reject(result); }
                    });

                    // log import duration
                    initGuiPromise.always(function () {
                        logLaunchDuration('postprocessing', 'launchEnd');
                    });
                });
            });

            // handle errors while initilizaling the application
            initDef.fail(function (message) {
                var cause = Utils.getStringOption(message, 'error', 'unknown');
                Notifications.yell('error', cause);
                Utils.error('BaseApplication.setLauncher(): initialization failed. Error code: "' + cause + '".');
                self.quit('error');
            });

            // Do not return a Deferred object that waits for the import process, application
            // must be registered synchronously, otherwise it is possible to open multiple
            // application instances for the same document.
        });

        // set quit callback invoked from the core method quit(), resolve/reject 'currentQuitDef'
        // according to the result of the registered before-quit handlers
        this.setQuit(function () {

            // call all before-quit handlers, rejecting one will resume application
            callHandlers(beforeQuitHandlers, quitReason)
            .done(function () {

                // Abort all running server requests, before invoking the quit handlers. Each
                // aborted request removes itself from the array, but a single abort() call
                // inside the while-loop MAY abort other dependent requests as well, therefore
                // a while loop will be used that checks the array length in each iteration.
                while (pendingRequests.length > 0) {
                    pendingRequests[0].abort();
                }

                // Bug 32989: Create a local copy of the quit handler array, and clear the
                // original array, before executing the quit handlers. This allows a quit
                // handler to register other special quit handlers that will be executed e.g.
                // when calling app.executeQuitHandlers() from logout while the quit handlers are
                // still running. Execute all quit handlers, do not care about the result.
                callHandlers(quitHandlers.splice(0), quitReason)
                .always(function () {

                    // fail-safety: quit handlers should not leave pending server requests
                    if (pendingRequests.length > 0) {
                        Utils.warn('BaseApplication.quit(): found pending server requests after quit handlers');
                        while (pendingRequests.length > 0) {
                            pendingRequests[0].abort();
                        }
                    }
                    pendingRequests = null;

                    // remove unload listener of the browser window
                    $(window).off('unload', unloadHandler);

                    // destroy singletons
                    _.each(singletons, function (singleton) {
                        if ((singleton instanceof BaseObject) && !singleton.destroyed) {
                            singleton.destroy();
                        }
                    });

                    // destroy MVC instances
                    controller.destroy();
                    view.destroy();
                    model.destroy();

                    // invoke all own destructor callbacks
                    _.each(destructors, function (destructor) { destructor.call(self); });

                    // bug 32948: 'currentQuitDef' must not be cleared to keep the application in 'quitting' state, needed in
                    // case another app.quit() will be called, e.g. from 'Reload document' after internal error while quitting
                    self = model = view = controller = destructors = importHandler = flushHandler = beforeQuitHandlers = quitHandlers = null;

                    // always clean custom URL hash params on quit
                    _.url.hash('standalone', null);
                    _.url.hash('closetimer', null);

                    // always resolve (really close the application), regardless
                    // of the result of the quit handlers
                    currentQuitDef.resolve();
                });

                // immediately prevent sending other server requests
                requestsLocked = true;
            })
            .fail(function () {
                // reject quit promise to indicate that the application remains alive
                currentQuitDef.reject();
            });

            return currentQuitDef.promise();
        });

        // prevent usage of these methods in derived classes
        delete this.setLauncher;
        delete this.setQuit;

    } // class BaseApplication

    // static methods ---------------------------------------------------------

    /**
     * Creates a static launcher that has to be returned by the main module of
     * each application type. The launcher tries to find a running application
     * which is working on a file described in the launch options passed to the
     * launcher. If no such application exists, it creates and returns a new
     * application object.
     *
     * @param {String} moduleName
     *  The application type identifier.
     *
     * @param {Function} ApplicationClass
     *  The constructor function of the application mix-in class that will
     *  extend the core application object. Receives the launch options passed
     *  to the launcher as first parameter.
     *
     * @param {Object} [appOptions]
     *  Optional parameters controlling the creation of the ox.ui.App base
     *  class instance:
     *  @param {String} [appOptions.icon]
     *      If specified, the CSS class name of a Bootstrap icon that will be
     *      shown in the top level launcher tab next to the application title.
     *  @param {Boolean} [options.search=false]
     *      Whether the application window will contain and support the global
     *      search tool bar.
     *  @param {Boolean} [options.chromeless=false]
     *      Whether to hide the main window tool bar attached to the left or
     *      bottom border.
     *
     * @returns {Object}
     *  The launcher object expected by the ox.launch() method.
     */
    BaseApplication.createLauncher = function (moduleName, ApplicationClass, appOptions) {

        var // the icon shown in the top bar launcher
            icon = Utils.getStringOption(appOptions, 'icon'),

            urlFileDescriptor = { id: _.url.hash('id'), folder_id: _.url.hash('folder') };

        // executed when a new application will be launched via ox.launch()
        function launchApp(launchOptions) {

            // Load from URL: if no launch options are given, check if we have 'simple' file descriptor (file id, folder id)
            // in the URL hash params and build new simple launch options from it
            if (!launchOptions && urlFileDescriptor.id && urlFileDescriptor.folder_id) {
                if (Config.getUrlFlag('convert')) {
                    launchOptions = {   actionSource: 'url',
                                        action: 'convert',
                                        file: null,
                                        folderId: urlFileDescriptor.folder_id,
                                        template: true,
                                        keepFile: true,
                                        initial_filename: _.url.hash('destfilename'),
                                        initial_folderId: _.url.hash('destfolderid'),
                                        templateFile: urlFileDescriptor
                                    };
                    _.url.hash('convert', null);
                    _.url.hash('destfilename', null);
                    _.url.hash('destfolderid', null);
                } else {
                    launchOptions = { actionSource: 'url', action: 'load', file: urlFileDescriptor};
                }
            }

            var // get file descriptor from options
                file = Utils.getObjectOption(launchOptions, 'file', null),
                // find running application for the specified document
                runningApps = file ? ox.ui.App.get(moduleName).filter(isRunningApplication) : [],
                // the new application instance
                newApp = null;

            // returns whether the passed application instance works on the exact
            // document file contained in the launch options
            function isRunningApplication(app) {
                var appFile = app.getFileDescriptor();
                return _.isObject(appFile) && Utils.hasEqualProperties(file, appFile, ['source', 'folder_id', 'id', 'attached']);
            }

            // return running application if existing
            if (runningApps.length > 1) {
                Utils.warn('BaseApplication.launchApp(): found multiple applications for the same file.');
            }
            if (runningApps.length > 0) {
                return runningApps[0];
            }

            // no running application: create and initialize a new application object
            newApp = ox.ui.createApp({
                name: moduleName,
                closable: true,
                userContent: icon.length > 0,
                userContentIcon: icon,
                userContentClass: getDocumentType(moduleName) + '-launcher'
            });

            // call mix-in constructor
            ApplicationClass.call(newApp, launchOptions, appOptions);

            return newApp;
        }

        // hide core AppSuite topbar in standalone mode
        if (getStandaloneMode()) {
            ext.point('io.ox/core/topbar').disable('default');
        }

        // build the complete CSS classes for the application icon
        icon = _.isString(icon) ? Forms.getIconClasses(icon) : '';

        // Bug 28664: remove all save points before they are checked to decide
        // whether to show the 'unsaved documents' dialog
        ext.point('io.ox/core/logout').extend({
            id: moduleName + '/logout/before',
            index: 'first', // run before the save points are checked by the core
            logout: function () {
                _.each(ox.ui.App.get(moduleName), function (app) {
                    if (!app.hasUnsavedChanges()) { app.removeRestorePoint(); }
                });
            }
        });

        // listen to user logout and notify all running applications
        ext.point('io.ox/core/logout').extend({
            id: moduleName + '/logout/quit',
            index: 'last', // run after the dialog (not called if logout was rejected)
            logout: function () {
                var promises = _.map(ox.ui.App.get(moduleName), function (app) {
                    // bug 33152: ignore the result of the quit handlers (rejected promise cancels logout)
                    var def = $.Deferred();
                    app.executeQuitHandlers('logout').always(function () { def.resolve(); });
                    return def.promise();
                });
                return $.when.apply($, promises);
            }
        });

        // ox.launch() expects an object with the method getApp()
        return { getApp: launchApp };
    };

    // exports ================================================================

    return _.makeExtendable(BaseApplication);

});
